API_SCHEDULER_ENABLE = 'api.scheduler.enable'
API_SCHEDULER_MUTE_LOGS = 'api.scheduler.mute-logs'
API_SCHEDULER_TIMEZONE = 'api.scheduler.time-zone'

API_SERVER_SCHEME = 'api.server.scheme'
API_SERVER_HOST = 'api.server.host'
API_SERVER_PORT = 'api.server.port'
API_SERVER_BASE_URL = 'api.server.base-url'

API_SESSION_SECRET = 'api.session.secret'
API_SESSION_ALGORITHM = 'api.session.algorithm'
API_SESSION_HEADER = 'api.session.header'
API_SESSION_TYPE = 'api.session.type'

API_API_KEY_SECRET = 'api.api-key.secret'
API_API_KEY_ALGORITHM = 'api.api-key.algorithm'
API_API_KEY_HEADER = 'api.api-key.header'
API_API_KEY_TYPE = 'api.api-key.type'

API_SECURITY_SECRET = 'api.security.secret'
API_SECURITY_ALGORITHM = 'api.security.algorithm'
API_SECURITY_HEADER = 'api.security.header'
API_SECURITY_TYPE = 'api.security.type'
