


class HeaderKey:
    CONTENT_TYPE = 'Content-Type'


class Verb:
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    PATCH = 'PATCH'
    OPTIONS = 'OPTIONS'

CONTROLLER_CONTEXT = 'ControllerRequest'
CLIENT_CONTEXT = 'ClientRequest'
